from typing import TYPE_CHECKING, Dict, List, Optional, Set, Union

from django.conf import settings
from rest_framework.settings import import_from_string, perform_import

if TYPE_CHECKING:
    from zq_django_util.utils.types import SimpleValue

    SettingValue = Union[
        None, SimpleValue, List["SettingValue"], Dict[str, "SettingValue"]
    ]
    SettingDict = Dict[str, SettingValue]


class PackageSettings:
    """
    Copy of DRF APISettings class with support for importing settings that
    are dicts with value as a string representing the path to the class
    to be imported.
    """

    setting_name: Optional[str] = None
    DEFAULTS: Optional["SettingDict"] = None
    IMPORT_STRINGS: Optional[List[str]] = None

    defaults: "SettingDict"
    import_strings: List[str]

    _cached_attrs: Set[str]
    _user_settings: "SettingDict"

    def __init__(
        self,
        defaults: Optional["SettingDict"] = None,
        import_strings: Optional[List[str]] = None,
    ):
        if self.setting_name is None:
            raise NotImplementedError("setting_name must be set")
        if self.DEFAULTS is None:
            raise NotImplementedError("DEFAULTS must be set")
        if self.IMPORT_STRINGS is None:
            raise NotImplementedError("IMPORT_STRINGS must be set")

        self.defaults = defaults or self.DEFAULTS
        self.import_strings = import_strings or self.IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self) -> "SettingDict":
        if not hasattr(self, "_user_settings"):
            assert self.setting_name is not None
            self._user_settings = getattr(settings, self.setting_name, {})
        return self._user_settings

    def __getattr__(self, attr: str) -> "SettingValue":
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            if isinstance(val, dict):
                val = {k: import_from_string(v, attr) for k, v in val.items()}
            else:
                val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self) -> None:
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")

    def reload_package_settings(self, *args, **kwargs) -> None:
        setting = kwargs["setting"]
        if setting == self.setting_name:
            self.reload()



# Generated by Django 4.1.4 on 2022-12-23 15:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="RequestLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "user",
                    models.IntegerField(
                        blank=True, null=True, verbose_name="用户ID"
                    ),
                ),
                ("ip", models.CharField(max_length=32, verbose_name="用户IP")),
                (
                    "method",
                    models.CharField(max_length=32, verbose_name="请求方法"),
                ),
                ("url", models.TextField(verbose_name="请求URL")),
                ("headers", models.JSONField(verbose_name="请求头")),
                (
                    "content_type",
                    models.CharField(max_length=32, verbose_name="请求类型"),
                ),
                ("query_param", models.JSONField(verbose_name="请求参数")),
                ("request_body", models.JSONField(verbose_name="请求数据")),
                ("file_data", models.JSONField(verbose_name="文件数据")),
                ("response", models.JSONField(verbose_name="响应数据")),
                (
                    "status_code",
                    models.PositiveSmallIntegerField(
                        db_index=True, verbose_name="响应状态码"
                    ),
                ),
                (
                    "execution_time",
                    models.DecimalField(
                        decimal_places=5,
                        max_digits=8,
                        null=True,
                        verbose_name="执行时间",
                    ),
                ),
                (
                    "create_time",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="请求时间"
                    ),
                ),
            ],
            options={
                "verbose_name": "请求日志",
                "verbose_name_plural": "请求日志",
                "db_table": "log_request",
                "ordering": ["-create_time"],
            },
        ),
        migrations.CreateModel(
            name="ExceptionLog",
            fields=[
                (
                    "requestlog_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="logs.requestlog",
                    ),
                ),
                (
                    "exp_id",
                    models.CharField(max_length=32, verbose_name="异常ID"),
                ),
                (
                    "exception_type",
                    models.CharField(max_length=128, verbose_name="异常类型"),
                ),
                (
                    "event_id",
                    models.CharField(max_length=32, verbose_name="Sentry事件ID"),
                ),
                ("exception_msg", models.TextField(verbose_name="异常信息")),
                ("exception_info", models.TextField(verbose_name="异常详情")),
                ("stack_info", models.JSONField(verbose_name="异常栈")),
            ],
            options={
                "verbose_name": "异常日志",
                "verbose_name_plural": "异常日志",
                "db_table": "log_exception",
                "ordering": ["-create_time"],
            },
            bases=("logs.requestlog",),
        ),
    ]



from typing import Optional, Union

import django.core.exceptions as django_exceptions
import rest_framework.exceptions as drf_exceptions
import rest_framework_simplejwt.exceptions as jwt_exceptions
from django.http import Http404
from drf_standardized_errors.formatter import ExceptionFormatter
from drf_standardized_errors.types import ExceptionHandlerContext
from rest_framework.response import Response
from rest_framework.views import set_rollback

from zq_django_util.exceptions import ApiException
from zq_django_util.exceptions.configs import zq_exception_settings
from zq_django_util.exceptions.types import ExtraHeaders
from zq_django_util.response import ResponseType
from zq_django_util.response.types import ApiExceptionResponse

if zq_exception_settings.SENTRY_ENABLE:  # pragma: no cover
    import sentry_sdk


class ApiExceptionHandler:
    exc: Exception
    context: ExceptionHandlerContext

    def __init__(
        self, exc: Exception, context: ExceptionHandlerContext
    ) -> None:
        self.exc = exc
        self.context = context

    def run(self) -> Optional[ApiExceptionResponse]:
        """
        处理异常
        :return: 响应数据或失败为None
        """
        exc = self.convert_known_exceptions(self.exc)  # 将django的异常转换为drf的异常

        if (
            zq_exception_settings.EXCEPTION_UNKNOWN_HANDLE
        ):  # 未知异常处理（非drf、api的异常）
            exc = self.convert_unhandled_exceptions(exc)  # 将未知异常转换为drf异常
        exc = self.convert_drf_exceptions(exc)  # 将drf异常转换为api异常
        set_rollback()  # 设置事务回滚
        response = None

        if isinstance(exc, ApiException):  # 如果是api异常则进行解析
            response = self.get_response(exc)
            if exc.record:  # 如果需要记录
                if zq_exception_settings.SENTRY_ENABLE:
                    self._notify_sentry(exc, response)
                # 将event_id写入响应数据
                response.data["data"]["event_id"] = exc.event_id
                # 将异常信息记录到response中，便于logger记录
                response.exception_data = exc

        return response

    def _notify_sentry(self, exc: ApiException, response: Response) -> None:
        """
        通知sentry, 可在notify_sentry中自定义其他内容
        :param exc: Api异常
        :param response: 响应数据
        :return: None
        """
        try:
            self.notify_sentry(exc, response)  # 调用自定义通知
        except Exception:
            pass

        # 默认异常汇报
        sentry_sdk.api.set_tag("exception_type", exc.response_type.name)
        sentry_sdk.set_context(
            "exp_info",
            {
                "eid": response.data["data"]["eid"],
                "code": response.data["code"],
                "detail": response.data["detail"],
                "msg": response.data["msg"],
            },
        )
        sentry_sdk.set_context("details", response.data["data"]["details"])
        exc.event_id = sentry_sdk.api.capture_exception(self.exc)  # 发送至sentry

    def notify_sentry(self, exc: ApiException, response: Response) -> None:
        """
        自定义sentry通知
        :param exc: Api异常
        :param response: 响应数据
        :return: None
        """
        user = self.context["request"].user
        if user.is_authenticated:
            sentry_sdk.api.set_tag("role", "user")
            sentry_sdk.api.set_user(
                {
                    "id": user.id,
                    "email": user.username,
                    "phone": user.phone if hasattr(user, "phone") else None,
                }
            )
        else:
            sentry_sdk.api.set_tag("role", "guest")

    def get_response(self, exc: ApiException) -> Response:
        """
        获取响应数据
        :param exc: Api异常
        :return:
        """
        headers = self.get_headers(exc.inner)
        data = exc.response_data

        info = None
        if exc.inner:
            if isinstance(exc.inner, drf_exceptions.APIException):
                info = ExceptionFormatter(
                    exc.inner, self.context, self.exc
                ).run()
            else:
                info = exc.inner.args[0] if exc.inner.args else None
        data["data"]["details"] = info

        return Response(
            data,
            status=exc.response_type.status_code,
            headers=headers,
        )

    @staticmethod
    def get_headers(exc: Exception) -> ExtraHeaders:
        """
        获取额外响应头
        :param exc: 异常
        :return: headers
        """
        headers = {}
        if getattr(exc, "auth_header", None):
            headers["WWW-Authenticate"] = exc.auth_header
        if getattr(exc, "wait", None):
            headers["Retry-After"] = "%d" % exc.wait
        return headers

    @staticmethod
    def convert_known_exceptions(exc: Exception) -> Exception:
        """
        By default, Django's built-in `Http404` and `PermissionDenied` are converted
        to their DRF equivalent.
        """
        if isinstance(exc, Http404):
            return drf_exceptions.NotFound()
        elif isinstance(exc, django_exceptions.PermissionDenied):
            return drf_exceptions.PermissionDenied()
        # jwt
        elif isinstance(exc, jwt_exceptions.InvalidToken):
            return ApiException(
                type=ResponseType.TokenInvalid,
                inner=exc,
            )
        elif isinstance(exc, jwt_exceptions.AuthenticationFailed):
            return ApiException(
                type=ResponseType.LoginFailed,
                inner=exc,
            )
        elif isinstance(exc, jwt_exceptions.TokenError):
            return ApiException(
                type=ResponseType.TokenInvalid,
                detail="Token解析错误",
                inner=exc,
            )
        else:
            return exc

    @staticmethod
    def convert_unhandled_exceptions(
        exc: Exception,
    ) -> Union[drf_exceptions.APIException, ApiException]:
        """
        Any non-DRF unhandled exception is converted to an APIException which
        has a 500 status code.
        """
        if not isinstance(exc, drf_exceptions.APIException) and not isinstance(
            exc, ApiException
        ):
            return drf_exceptions.APIException(detail=str(exc))
        else:
            return exc

    @staticmethod
    def convert_drf_exceptions(
        exc: Union[drf_exceptions.APIException, ApiException, Exception],
    ) -> ApiException | Exception:
        """
        转换drf异常
        :param exc: drf异常
        :return: ApiException
        """
        if isinstance(exc, ApiException):
            return exc
        # 处理drf异常
        record = False
        response_type = ResponseType.ServerError
        detail = None
        msg = None

        if isinstance(exc, drf_exceptions.ParseError):
            response_type = ResponseType.JSONParseFailed
        elif isinstance(exc, drf_exceptions.AuthenticationFailed):
            response_type = ResponseType.LoginFailed
        elif isinstance(exc, drf_exceptions.NotAuthenticated):
            # 未登录
            response_type = ResponseType.NotLogin
        elif isinstance(exc, drf_exceptions.PermissionDenied):
            response_type = ResponseType.PermissionDenied
        elif isinstance(exc, drf_exceptions.NotFound):
            response_type = ResponseType.APINotFound
        elif isinstance(exc, drf_exceptions.ValidationError):
            # 校验失败
            record = True
            response_type = ResponseType.ParamValidationFailed
        elif isinstance(exc, drf_exceptions.MethodNotAllowed):
            # 方法错误
            record = True
            response_type = ResponseType.MethodNotAllowed
            detail = f"不允许{getattr(exc, 'args', None)[0]}请求"
        elif isinstance(exc, drf_exceptions.NotAcceptable):
            record = True
            response_type = ResponseType.HeaderNotAcceptable
            detail = f"不支持{getattr(exc, 'args', None)[0]}的响应格式"
        elif isinstance(exc, drf_exceptions.UnsupportedMediaType):
            record = True
            response_type = ResponseType.UnsupportedMediaType
            detail = f"不支持{getattr(exc, 'args', None)[0]}的请求格式"
            msg = f"暂不支持{getattr(exc, 'args', None)[0]}文件上传，请使用支持的文件格式重试"
        elif isinstance(exc, drf_exceptions.Throttled):
            record = True
            response_type = ResponseType.APIThrottled
            detail = f"请求频率过高，请{getattr(exc, 'args', None)[0]}s后再试"
            msg = f"请求太快了，请{getattr(exc, 'args', None)[0]}s后再试"
        elif isinstance(exc, drf_exceptions.APIException):
            record = True
            response_type = ResponseType.ServerError
        else:
            # 不处理其他异常
            return exc

        return ApiException(response_type, msg, exc, record, detail)


def exception_handler(
    exc: Exception, context: ExceptionHandlerContext
) -> Optional[ApiExceptionResponse]:
    """
    自定义异常处理

    :param exc: 异常
    :param context: 上下文
    :return: 处理程序
    """
    handler_class = zq_exception_settings.EXCEPTION_HANDLER_CLASS

    if handler_class != ApiExceptionHandler and not issubclass(
        handler_class, ApiExceptionHandler
    ):
        raise ImportError(
            f"{handler_class} is not a subclass of ApiExceptionHandler"
        )

    return handler_class(exc, context).run()



import re
from typing import Dict

from django.db import close_old_connections
from rest_framework.request import Request

from zq_django_util.logs.configs import drf_logger_settings
from zq_django_util.response.types import JSONVal


def get_headers(request: Request = None) -> Dict[str, str]:
    """
    Function:       get_headers(self, request)
    Description:    To get all the headers from request
    """
    regex = re.compile("^HTTP_")
    return dict(
        (regex.sub("", header), value)
        for (header, value) in request.META.items()
        if header.startswith("HTTP_")
    )


def get_client_ip(request: Request) -> str:
    try:
        x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
        if x_forwarded_for:
            ip = x_forwarded_for.split(",")[0]
        else:
            ip = request.META.get("REMOTE_ADDR")
        return ip
    except Exception:
        return ""


def is_api_logger_enabled() -> bool:
    return drf_logger_settings.DATABASE or drf_logger_settings.SIGNAL


def database_log_enabled() -> bool:
    return drf_logger_settings.DATABASE


def mask_sensitive_data(data: JSONVal) -> JSONVal:
    """
    Hides sensitive keys specified in sensitive_keys settings.
    Loops recursively over nested dictionaries.
    """

    if type(data) != dict:
        return data

    for key, value in data.items():
        if key in drf_logger_settings.SENSITIVE_KEYS:
            length = len(data[key])
            data[key] = f"***FILTERED*** (len: {length})"

        if type(value) == dict:
            data[key] = mask_sensitive_data(data[key])

        if type(value) == list:
            data[key] = [mask_sensitive_data(item) for item in data[key]]

    return data


def close_old_database_connections(func):
    """
    自定义decorator，用来装饰使用数据库操作函数

    https://stackoverflow.com/questions/59773675/why-am-i-getting-the-mysql-server-has-gone-away-exception-in-django
    """

    def wrapper(*args, **kwargs):
        close_old_connections()

        return func(*args, **kwargs)

    return wrapper



import base64
import datetime
import hashlib
import hmac
import json
import random
import time
from typing import AnyStr, Dict, Optional, Tuple, TypedDict
from urllib.parse import unquote
from urllib.request import urlopen

from Crypto.Hash import MD5
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from django.core.cache import cache
from rest_framework.request import Request

from zq_django_util.utils.oss.configs import oss_settings
from zq_django_util.utils.types import JSONValue


class OSSCallbackToken(TypedDict):
    OSSAccessKeyId: str
    host: str
    policy: str
    signature: str
    expire: str
    key: str
    callback: str


def get_iso_8601(expire: float) -> str:
    gmt = datetime.datetime.utcfromtimestamp(expire).isoformat()
    gmt += "Z"
    return gmt


def split_file_name(file_name: str) -> Tuple[str, str]:
    """
    获取文件名与扩展名

    :param file_name: 文件全名

    :return: 文件名，扩展名
    """
    if "." in file_name:  # 文件存在扩展名
        ext = file_name.split(".")[-1]  # 文件扩展名
        name = ".".join(file_name.split(".")[0:-1])
    else:
        ext = ""
        name = file_name

    return name, ext


def get_random_name(file_name: str) -> str:
    """
    获取随机文件名

    :param file_name: 原文件名

    :return:
    """
    name, ext = split_file_name(file_name)

    new_name = time.strftime("%Y%m%d%H%M%S")  # 定义文件名，年月日时分秒随机数
    new_name = (
        new_name
        + "_%04d" % random.randint(0, 10000)
        + (("." + ext) if ext != "" else "")
    )

    return new_name


def get_token(
    key: str,
    callback: Dict[str, str],
    policy: Optional[JSONValue] = None,
) -> OSSCallbackToken:
    """
    获取直传签名token
    """
    # 获取Policy
    expire_time = datetime.datetime.now() + datetime.timedelta(
        seconds=oss_settings.TOKEN_EXPIRE_SECOND
    )
    expire = get_iso_8601(expire_time.timestamp())
    if policy is None:
        policy = {
            "expiration": expire,  # 过期时间
            "conditions": [
                {"bucket": oss_settings.BUCKET_NAME},
                [
                    "content-length-range",
                    0,
                    oss_settings.MAX_SIZE_MB * 1024 * 1024,
                ],  # 限制上传文件大小
                ["eq", "$key", f"{key}"],  # 限制上传文件名
            ],
        }
    policy = json.dumps(policy).strip()
    policy_encode = base64.b64encode(policy.encode())

    # 签名
    h = hmac.new(
        oss_settings.ACCESS_KEY_SECRET.encode(),
        policy_encode,
        hashlib.sha1,
    )
    sign = base64.encodebytes(h.digest()).strip()

    # 回调参数
    callback_param = json.dumps(callback).strip()
    base64_callback_body = base64.b64encode(callback_param.encode())

    return dict(
        OSSAccessKeyId=oss_settings.ACCESS_KEY_ID,
        host=(
            f'{oss_settings.ENDPOINT.split("://")[0]}://{oss_settings.BUCKET_NAME}.'
            f'{oss_settings.ENDPOINT.split("://")[1]}'
        ),
        policy=policy_encode.decode(),
        signature=sign.decode(),
        expire=expire,
        key=key,
        callback=base64_callback_body.decode(),
    )


def check_callback_signature(request: Request) -> bool:
    """
    检测回调身份
    """
    authorization_base64 = request.META.get(
        "HTTP_AUTHORIZATION", None
    )  # 获取AUTHORIZATION
    pub_key_url_base64 = request.META.get(
        "HTTP_X_OSS_PUB_KEY_URL", None
    )  # 获取公钥
    if authorization_base64 is None or pub_key_url_base64 is None:
        return False

    try:
        # 对x-oss-pub-key-url做base64解码后获取到公钥
        pub_key_url = base64.b64decode(pub_key_url_base64).decode()

        # 为了保证该public_key是由OSS颁发的，用户需要校验x-oss-pub-key-url的开头
        if not pub_key_url.startswith(
            "http://gosspublic.alicdn.com/"
        ) and not pub_key_url.startswith("https://gosspublic.alicdn.com/"):
            return False
        pub_key = get_pub_key(pub_key_url)

        # 获取base64解码后的签名
        authorization = base64.b64decode(authorization_base64)

        # 获取待签名字符串
        callback_body = request.body

        if request.META["QUERY_STRING"] == "":
            auth_str = (
                unquote(request.META["PATH_INFO"])
                + "\n"
                + callback_body.decode()
            )
        else:
            auth_str = (
                unquote(request.META["PATH_INFO"])
                + "?"
                + request.META["QUERY_STRING"]
                + "\n"
                + callback_body.decode()
            )

        # 验证签名
        auth_md5 = MD5.new(auth_str.encode())
        rsa_pub = RSA.importKey(pub_key)
        verifier = PKCS1_v1_5.new(rsa_pub)
        verifier.verify(auth_md5, authorization)
        return True
    except Exception:
        return False


def _get_pub_key_online(pub_key_url: str) -> AnyStr:
    """
    从网络获取pub key
    """
    response = urlopen(pub_key_url)
    return response.read()


def get_pub_key(pub_key_url: str) -> AnyStr:
    """
    获取公钥
    :param pub_key_url: url
    :return:
    """
    key = f"oss:pub_key:{pub_key_url}"

    try:
        res = cache.get(key, None)
        if res is None:
            res = _get_pub_key_online(pub_key_url)
            cache.set(key, res)
        return res
    except Exception:
        return _get_pub_key_online(pub_key_url)



from rest_framework.pagination import PageNumberPagination


class GlobalPageNumberPagination(PageNumberPagination):
    """
    自定义分页
    """

    def __init__(self):
        super(GlobalPageNumberPagination, self).__init__()
        # TODO config化参数，嵌套类进行参数局部定义
        self.page_size_query_param = "page_size"
        self.page_size = 20
        self.max_page_size = 200



from datetime import datetime
from typing import Dict, List, Optional, TypedDict, Union

from zq_django_util.response.types import JSONVal

HeaderContent = Union[str, List["HeaderContent"], Dict[str, "HeaderContent"]]
HeaderDict = Dict[str, HeaderContent]


class FileDataDict(TypedDict, total=True):
    name: str
    size: int
    content_type: str
    content_type_extra: str


class RequestLogDict(TypedDict, total=True):
    user: Optional[int]
    ip: str
    method: str
    url: str
    headers: HeaderDict
    content_type: str
    query_param: Dict[str, JSONVal]
    request_body: Dict[str, JSONVal]
    file_data: Dict[str, FileDataDict]
    response: JSONVal
    status_code: int
    execution_time: Optional[float]
    create_time: datetime


class ExceptionLogDict(RequestLogDict, total=True):
    exp_id: Optional[str]
    event_id: Optional[str]
    exception_type: str
    exception_msg: str
    exception_info: str
    stack_info: List[str]



from typing import List, TypedDict

from django.conf import settings
from django.core.signals import setting_changed
from django.dispatch import receiver

from zq_django_util.utils.package_settings import PackageSettings

ZqExceptionSettingDict = TypedDict(
    "ZqExceptionSettingDict",
    {
        "EXCEPTION_UNKNOWN_HANDLE": bool,
        "EXCEPTION_HANDLER_CLASS": str,
    },
    total=True,
)


class ZqExceptionSettings(PackageSettings):
    setting_name = "ZQ_EXCEPTION"

    DEFAULTS: ZqExceptionSettingDict = {
        "EXCEPTION_UNKNOWN_HANDLE": True,  # 处理未知异常
        "EXCEPTION_HANDLER_CLASS": "zq_django_util.exceptions.handler.ApiExceptionHandler",
    }

    IMPORT_STRINGS: List[str] = ["EXCEPTION_HANDLER_CLASS"]

    SENTRY_ENABLE: bool = getattr(settings, "SENTRY_ENABLE", False)


zq_exception_settings = ZqExceptionSettings()


@receiver(setting_changed)
def reload_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "SENTRY_ENABLE":
        zq_exception_settings.SENTRY_ENABLE = getattr(
            settings, "SENTRY_ENABLE", False
        )
    zq_exception_settings.reload_package_settings(*args, **kwargs)



import hashlib
import random
import time
import traceback
from datetime import datetime
from sys import exc_info
from typing import TYPE_CHECKING, Optional

from django.conf import settings
from django.utils.timezone import now

if TYPE_CHECKING:
    from zq_django_util.exceptions.types import ExceptionData, ExceptionInfo
    from zq_django_util.response import ResponseData, ResponseType


class ApiException(Exception):
    """API异常"""

    DEFAULT_MSG = "服务器开小差了"
    RECORD_MSG_TEMPLATE = "请向工作人员反馈以下内容："

    record: bool
    response_type: "ResponseType"
    eid: Optional[str]
    detail: str
    msg: str
    event_id: Optional[str]
    inner: Optional[Exception]
    _exc_data: Optional["ExceptionInfo"]
    time: datetime

    def __init__(
        self,
        type: "ResponseType",
        msg: Optional[str] = None,
        inner: Optional[Exception] = None,
        record: Optional[bool] = None,
        detail: Optional[str] = None,
    ) -> None:
        """
        API异常
        :param type: 异常类型
        :param msg: 异常建议(面向用户)
        :param inner: 内部异常
        :param record: 是否记录异常
        :param detail: 异常详情(面向开发者)
        """
        self.record = (
            record if record is not None else type.status_code == 500
        )  # 是否记录异常

        self.response_type: "ResponseType" = type
        if self.record:  # 记录异常
            self.eid = self.get_exp_id()  # 异常id
        else:
            self.eid = None
        self.detail = self.get_exp_detail(detail)  # 异常详情
        self.msg = self.get_exp_msg(msg)  # 异常用户提示
        self.event_id = None

        super().__init__(self.detail)

        self.inner = inner
        self._exc_data = None
        self.time = now()

    def get_exp_detail(self, detail: Optional[str]) -> str:
        """
        获取异常详情(面向开发者)
        :param detail: 自定义异常详情
        :return: 异常详情
        """
        res = detail or self.response_type.detail  # 获取异常详情
        if self.record:  # 记录异常
            res = f"{res}, {self.eid}"  # 将自定义异常详情添加到末尾
        return res

    def get_exp_msg(self, msg: Optional[str]) -> str:
        """
        获取异常用户提示(面向用户)
        :param msg: 自定义异常用户提示
        :return: 异常用户提示
        """
        if self.record:  # 记录异常
            if msg:  # 如果有自定义异常用户提示
                res = f"{msg}，{self.RECORD_MSG_TEMPLATE}{self.eid}"  # 使用自定义用户提示
            else:
                res = f"{self.DEFAULT_MSG}，{self.RECORD_MSG_TEMPLATE}{self.eid}"  # 标准用户提示
        else:  # 不记录异常
            res = msg or self.response_type.detail  # 获取异常详情
        return res

    @property
    def exc_data(self) -> "ExceptionInfo":
        if self._exc_data is None:
            self._exc_data = self.get_exception_info()
        return self._exc_data

    @exc_data.setter
    def exc_data(self, value: "ExceptionInfo") -> None:
        self._exc_data = value

    @property
    def response_data(self) -> "ResponseData":
        """
        获取响应数据
        :return: 响应数据
        """
        return {
            "code": self.response_type.code,
            "detail": self.detail,
            "msg": self.msg,
            "data": self.exception_data,
        }

    @staticmethod
    def get_exp_id() -> str:
        """
        获取异常id
        :return: 异常id
        """
        sha = hashlib.sha1()
        exp_id = time.strftime("%Y%m%d%H%M%S") + "_%04d" % random.randint(
            0, 10000
        )
        sha.update(exp_id.encode("utf-8"))
        return sha.hexdigest()[:6]

    @staticmethod
    def get_exception_info() -> "ExceptionInfo":
        """
        获取异常信息
        :return: 异常信息
        """
        exc_type, exc_value, exc_traceback_obj = exc_info()
        return {
            "type": str(exc_type),
            "msg": str(exc_value),
            "info": traceback.format_exc(),
            "stack": traceback.format_stack(),
        }

    @property
    def exception_data(self) -> "ExceptionData":
        """
        获取异常返回数据
        :return: 返回数据
        """
        data: "ExceptionData" = {
            "eid": self.eid,
            "time": self.time,
        }

        if settings.DEBUG:
            data["exception"] = self.exc_data

        return data



from dataclasses import dataclass
from enum import Enum, unique
from typing import TYPE_CHECKING, Optional

from django.http import JsonResponse

if TYPE_CHECKING:
    from zq_django_util.exceptions import ApiException
    from zq_django_util.response.types import JSONVal, ResponseData


class ResponseTypeEnum(Enum):
    @property
    def code(self) -> str:
        """
        根据枚举名称取状态码code

        :return: 状态码code
        """
        return self.value[0]

    @property
    def detail(self) -> str:
        """
        根据枚举名称取状态说明message

        :return: 状态说明message
        """
        return self.value[1]

    @property
    def status_code(self) -> int:
        """
        根据枚举名称取状态码status_code

        :return: 状态码status_code
        """
        return self.value[2]


# region ResponseType
@unique
class ResponseType(ResponseTypeEnum):
    """API状态类型"""

    Success = ("00000", "", 200)
    ClientError = ("A0000", "用户端错误", 400)
    LoginFailed = ("A0210", "用户登录失败", 400)
    UsernameNotExist = ("A0211", "用户名不存在", 400)
    PasswordWrong = ("A0212", "用户密码错误", 400)
    LoginFailedExceed = ("A0213", "用户输入密码次数超限", 400)
    PhoneNotExist = ("A0214", "手机号不存在", 400)
    LoginExpired = ("A0220", "用户登录已过期", 401)
    TokenInvalid = ("A0221", "token 无效或已过期", 401)
    RefreshTokenInvalid = ("A0221", "refresh token 无效或已过期", 401)
    ThirdLoginFailed = ("A0230", "用户第三方登录失败", 401)
    ThirdLoginCaptchaError = ("A0232", "用户第三方登录验证码错误", 401)
    ThirdLoginExpired = ("A0233", "用户第三方登录已过期", 401)
    PermissionError = ("A0300", "用户权限异常", 403)
    NotLogin = ("A0310", "用户未登录", 401)
    NotActive = ("A0311", "用户未激活", 403)
    PermissionDenied = ("A0312", "用户无权限", 403)
    ServiceNotAvailable = ("A0313", "不在服务时段", 403)
    UserBlocked = ("A0320", "黑名单用户", 403)
    UserFrozen = ("A0321", "账号被冻结", 403)
    IPInvalid = ("A0322", "非法 IP 地址", 401)
    ParamError = ("A0400", "用户请求参数错误", 400)
    JSONParseFailed = ("A0410", "请求 JSON 解析错误", 400)
    ParamEmpty = ("A0420", "请求必填参数为空", 400)
    ParamValidationFailed = ("A0430", "请求参数值校验失败", 400)
    RequestError = ("A0500", "用户请求服务异常", 400)
    APINotFound = ("A0510", "请求接口不存在", 404)
    MethodNotAllowed = ("A0511", "请求方法不允许", 405)
    APIThrottled = ("A0512", "请求次数超出限制", 429)
    HeaderNotAcceptable = ("A0513", "请求头无法满足", 406)
    ResourceNotFound = ("A0514", "请求资源不存在", 404)
    UploadError = ("A0600", "用户上传文件异常", 400)
    UnsupportedMediaType = ("A0610", "用户上传文件类型不支持", 400)
    UnsupportedMediaSize = ("A0613", "用户上传文件大小错误", 400)
    VersionError = ("A0700", "用户版本异常", 400)
    AppVersionError = ("A0710", "用户应用安装版本不匹配", 400)
    APIVersionError = ("A0720", "用户 API 请求版本不匹配", 400)
    ServerError = ("B0000", "系统执行出错", 500)
    ServerTimeout = ("B0100", "系统执行超时", 500)
    ServerResourceError = ("B0200", "系统资源异常", 500)
    ThirdServiceError = ("C0000", "调用第三方服务出错", 500)
    MiddlewareError = ("C0100", "中间件服务出错", 500)
    ThirdServiceTimeoutError = ("C0200", "第三方系统执行超时", 500)
    DatabaseError = ("C0300", "数据库服务出错", 500)
    CacheError = ("C0400", "缓存服务出错", 500)
    NotificationError = ("C0500", "通知服务出错", 500)


# endregion


@dataclass
class ApiResponse:
    """API响应数据结构"""

    status_code: int
    code: str
    detail: str
    msg: str
    data: "ResponseData"

    def __init__(
        self,
        response_type: ResponseType = ResponseType.Success,
        data: "JSONVal" = "",
        msg: Optional[str] = None,
        ex: Optional["ApiException"] = None,
    ):
        """
        Api 响应
        :param response_type: 响应类型
        :param data: 响应数据
        :param msg: 面向用户的响应消息
        :param ex: Api异常(用于获取相关数据)
        """
        if ex:  # 优先使用异常
            response_type = ex.response_type  # 获取传入异常的类型
            msg = ex.msg  # 获取传入异常的消息

        self.status_code = response_type.status_code
        self.code = response_type.code
        self.detail = response_type.detail
        self.msg = msg if msg else self.detail
        self.data = data

    def __str__(self) -> str:
        return f"code: {self.code}, detail: {self.detail}, msg: {self.msg}, data: {self.data}"

    def __dict__(self) -> "ResponseData":
        return {
            "code": self.code,
            "detail": self.detail,
            "msg": self.msg,
            "data": self.data,
        }

    def to_json_response(self) -> JsonResponse:
        return JsonResponse(self.__dict__(), status=self.status_code)



from typing import List, Optional, TypedDict

from django.core.signals import setting_changed
from django.dispatch import receiver

from zq_django_util.utils.package_settings import PackageSettings

DrfLoggerSettingDict = TypedDict(
    "DrfLoggerSettingDict",
    {
        "DEFAULT_DATABASE": str,
        "QUEUE_MAX_SIZE": int,
        "INTERVAL": int,
        "DATABASE": bool,
        "SIGNAL": bool,
        "PATH_TYPE": str,
        "SKIP_URL_NAME": List[str],
        "SKIP_NAMESPACE": List[str],
        "METHODS": Optional[List[str]],
        "STATUS_CODES": Optional[List[int]],
        "SENSITIVE_KEYS": List[str],
        "ADMIN_SLOW_API_ABOVE": int,  # ms
        "ADMIN_TIMEDELTA": int,  # minute
    },
    total=True,
)


class DrfLoggerSettings(PackageSettings):
    setting_name = "DRF_LOGGER"

    DEFAULTS: DrfLoggerSettingDict = {
        "DEFAULT_DATABASE": "default",
        "QUEUE_MAX_SIZE": 50,
        "INTERVAL": 10,
        "DATABASE": False,
        "SIGNAL": False,
        "PATH_TYPE": "FULL_PATH",
        "SKIP_URL_NAME": [],
        "SKIP_NAMESPACE": [],
        "METHODS": None,
        "STATUS_CODES": None,
        "SENSITIVE_KEYS": ["password", "token", "access", "refresh"],
        "ADMIN_SLOW_API_ABOVE": 500,
        "ADMIN_TIMEDELTA": 0,
    }

    IMPORT_STRINGS: List[str] = []


drf_logger_settings = DrfLoggerSettings()


@receiver(setting_changed)
def reload_settings(*args, **kwargs):
    drf_logger_settings.reload_package_settings(*args, **kwargs)



from zq_django_util.logs.utils import is_api_logger_enabled

LOGGER_THREAD = None

if is_api_logger_enabled():
    import threading

    from zq_django_util.logs.handler import HandleLogAsync

    LOG_THREAD_NAME = "log_thread"

    already_exists = False

    for t in threading.enumerate():
        if t.name == LOG_THREAD_NAME:
            already_exists = True
            break

    if not already_exists:
        t = HandleLogAsync()
        t.daemon = True
        t.name = LOG_THREAD_NAME
        t.start()
        LOGGER_THREAD = t



from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from rest_framework.response import Response

if TYPE_CHECKING:
    from rest_framework.request import Request
    from rest_framework.viewsets import GenericViewSet

CACHE_TTL = getattr(settings, "CACHE_TTL", 60 * 60 * 1)


class CacheListModelMixin:
    @method_decorator(cache_page(CACHE_TTL))
    def list(
        self: "GenericViewSet", request: "Request", *args: Any, **kwargs: Any
    ) -> Response:
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)


class CacheRetrieveModelMixin:
    @method_decorator(cache_page(CACHE_TTL))
    def retrieve(
        self: "GenericViewSet", request: "Request", *args: Any, **kwargs: Any
    ) -> Response:
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)



from datetime import datetime
from typing import List, Optional, TypedDict


class ExceptionInfo(TypedDict, total=True):
    type: str
    msg: str
    info: str
    stack: List[str]


class ExceptionData(TypedDict, total=False):
    eid: Optional[str]
    time: datetime
    exception: Optional[ExceptionInfo]


ExtraHeaders = TypedDict(
    "ExtraHeaders",
    {
        "WWW-Authenticate": str,
        "Retry-After": str,
    },
    total=False,
)



from typing import TYPE_CHECKING, Any

from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.response import Response
from rest_framework_simplejwt.exceptions import TokenError
from rest_framework_simplejwt.views import TokenObtainPairView

from zq_django_util.exceptions import ApiException
from zq_django_util.response import ResponseType
from zq_django_util.utils.auth.serializers import (
    OpenIdLoginSerializer,
    PasswordLoginSerializer,
)

if TYPE_CHECKING:
    from rest_framework.request import Request

AuthUser = get_user_model()


class OpenIdLoginView(TokenObtainPairView):
    """
    open id 登录视图（仅供测试微信登录使用）
    """

    queryset = AuthUser.objects.all()
    serializer_class = OpenIdLoginSerializer

    def post(self, request: "Request", *args: Any, **kwargs: Any) -> Response:
        """
        增加 post 方法, 支持 open id 登录
        """
        serializer = self.get_serializer(data=request.data)

        try:
            serializer.is_valid(raise_exception=True)
        except TokenError:
            raise ApiException(
                ResponseType.ThirdLoginFailed,
                msg="微信登录失败",
                detail="生成token时simple jwt发生TokenError",
                record=True,
            )

        return Response(serializer.validated_data, status=status.HTTP_200_OK)


class PasswordLoginView(TokenObtainPairView):
    """
    密码登录视图
    """

    serializer_class = PasswordLoginSerializer



from typing import Dict, List, TypedDict, Union

from rest_framework.response import Response

from zq_django_util.exceptions import ApiException

JSONVal = Union[
    None, bool, str, float, int, List["JSONVal"], Dict[str, "JSONVal"]
]


class ResponseData(TypedDict, total=True):
    code: str
    detail: str
    msg: str
    data: JSONVal


class ApiExceptionResponse(Response):
    api_request_data: Dict[str, JSONVal]
    exception_data: ApiException
    exception: bool



from typing import Any

from django.utils import timezone
from rest_framework.mixins import ListModelMixin
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet


class APIRootViewSet(ListModelMixin, GenericViewSet):
    """
    API root view.
    """

    def list(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        user = request.user
        time = timezone.localtime().isoformat()

        if user.is_anonymous:
            return Response(
                dict(
                    user=None,
                    time=time,
                )
            )
        else:
            return Response(
                dict(
                    user=self.handle_user_info(user),
                    time=time,
                )
            )

    def handle_user_info(self, user) -> dict:
        """
        处理用户信息
        :param user: 用户对象
        :return: 返回用户信息
        """
        return dict(
            id=user.id,
            username=user.username,
            is_active=user.is_active,
            is_superuser=user.is_superuser,
        )



import base64
import json
import time
from logging import getLogger
from queue import Queue
from threading import Thread
from typing import Dict, List, Optional

from django.core.files.uploadedfile import UploadedFile
from django.db.utils import OperationalError
from django.urls import resolve
from django.utils import timezone
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework_simplejwt.settings import api_settings

import zq_django_util
from zq_django_util.exceptions import ApiException
from zq_django_util.logs.configs import drf_logger_settings
from zq_django_util.logs.models import ExceptionLog, RequestLog
from zq_django_util.logs.types import (
    ExceptionLogDict,
    FileDataDict,
    RequestLogDict,
)
from zq_django_util.logs.utils import (
    close_old_database_connections,
    get_client_ip,
    get_headers,
    mask_sensitive_data,
)
from zq_django_util.response.types import ApiExceptionResponse, JSONVal

logger = getLogger("drf_logger")


class HandleLogAsync(Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flag = True
        self._queue: Queue[(Request, Response, float, float)] = Queue(
            maxsize=drf_logger_settings.QUEUE_MAX_SIZE
        )

    def __enter__(self):
        self.daemon = True
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def run(self) -> None:
        """
        线程开始
        :return:
        """
        self.flag = True
        self.start_queue_process()

    def stop(self) -> None:
        """
        强制结束线程
        :return:
        """
        self.flag = False
        self.join()

    def prepare_request_log(
        self,
        request: Request,
        response: ApiExceptionResponse,
        start_time: float,
        end_time: float | None,
    ) -> Optional[RequestLogDict]:
        """
        处理请求日志
        :param request:
        :param response:
        :param start_time:
        :param end_time:
        :return:
        """
        # region 检查是否需要记录日志
        if not drf_logger_settings.DATABASE and not drf_logger_settings.SIGNAL:
            return

        url_name = resolve(request.path_info).url_name
        namespace = resolve(request.path_info).namespace

        # Always skip Admin panel
        if (
            namespace == "admin"
            or namespace == "__debug__"
            or url_name in drf_logger_settings.SKIP_URL_NAME
            or namespace in drf_logger_settings.SKIP_NAMESPACE
        ):
            return

        # Only log required status codes if matching
        if (
            drf_logger_settings.STATUS_CODES is not None
            and response.status_code not in drf_logger_settings.STATUS_CODES
        ):
            return

        # Log only registered methods if available.
        if (
            drf_logger_settings.METHODS is not None
            and request.method not in drf_logger_settings.METHODS
        ):
            return

        # endregion
        data = self.get_request_log_data(
            request, response, start_time, end_time
        )  # 解析数据

        if drf_logger_settings.SIGNAL:  # 需要发送信号
            # TODO 使用django信号发送日志
            pass

        if drf_logger_settings.DATABASE:  # 需要写入数据库
            return data  # 返回数据

    def put_log_data(
        self,
        request: Request,
        response: Response,
        start_time: float,
        end_time: float | None = None,
    ) -> None:
        """
        将日志数据放入队列
        :param request:
        :param response:
        :param start_time:
        :param end_time:
        :return:
        """
        self._queue.put((request, response, start_time, end_time))

        if self._queue.qsize() >= drf_logger_settings.QUEUE_MAX_SIZE:
            # 队列已满，开始处理日志
            self._start_log_parse()

    def start_queue_process(self):
        """
        持续处理日志
        :return:
        """
        while self.flag:
            time.sleep(drf_logger_settings.INTERVAL)  # 睡眠
            if self.flag:
                self._start_log_parse()

    def _start_log_parse(self) -> None:
        """
        开始处理日志
        :return:
        """
        request_items: List[RequestLog] = []  # 请求日志
        exception_items: List[ExceptionLog] = []  # 异常日志
        while not self._queue.empty():
            try:
                request: Request
                response: Response | ApiExceptionResponse
                start_time: float
                end_time: float | None

                request, response, start_time, end_time = self._queue.get()

                # 存在异常信息，则为异常日志
                if getattr(response, "exception_data", False):
                    res = self.prepare_exception_log(
                        request, response, start_time, end_time
                    )
                    if res:  # 解析后需要插入数据库
                        exception_items.append(ExceptionLog(**res))
                else:  # 否则只记录请求日志
                    res = self.prepare_request_log(
                        request, response, start_time, end_time
                    )
                    if res:  # 解析后需要插入数据库
                        request_items.append(RequestLog(**res))
            except Exception:
                pass

        if request_items or exception_items:  # 有日志需要写入数据库
            self._insert_into_database(request_items, exception_items)

    @staticmethod
    @close_old_database_connections
    def _insert_into_database(
        request_items: List[RequestLog],
        exception_items: List[ExceptionLog],
    ) -> None:
        """
        写入数据库
        :param request_items: 请求日志列表
        :param exception_items: 异常日志列表
        :return:
        """
        try:
            if request_items:  # 有请求日志
                zq_django_util.logs.models.RequestLog.objects.using(
                    drf_logger_settings.DEFAULT_DATABASE
                ).bulk_create(
                    request_items
                )  # 批量插入
                logger.debug(
                    f"insert {len(request_items)} request log into database"
                )
            if exception_items:  # 有异常日志
                # TODO 无法直接使用bulk_create: Can't bulk create a multi-table inherited model
                for item in exception_items:  # 逐条插入
                    item.save(using=drf_logger_settings.DEFAULT_DATABASE)
                logger.debug(
                    f"insert {len(exception_items)} exception log into database"
                )
        except OperationalError:  # 没有相关数据库表
            raise Exception(
                """
            DRF API LOGGER EXCEPTION
            Model does not exists.
            Did you forget to migrate?
            """
            )
        except Exception as e:  # 其他异常
            logger.error(f"DRF API LOGGER EXCEPTION: {e}")

    @classmethod
    def prepare_exception_log(
        cls,
        request: Request,
        response: ApiExceptionResponse,
        start_time: float,
        end_time: float | None,
    ) -> ExceptionLogDict:
        """
        解析异常记录
        :param request: 请求
        :param response: 响应
        :param start_time: 开始时间
        :param end_time: 结束时间
        :return: 异常数据
        """
        data: RequestLogDict = cls.get_request_log_data(
            request, response, start_time, end_time
        )  # 获取请求日志数据
        exception_data: ApiException = response.exception_data
        data.update(
            dict(
                exp_id=exception_data.eid or "",
                event_id=exception_data.event_id or "",
                exception_type=exception_data.exc_data["type"],
                exception_msg=exception_data.exc_data["msg"],
                exception_info=exception_data.exc_data["info"],
                stack_info=exception_data.exc_data["stack"],
            )
        )
        return data

    @staticmethod
    def get_request_log_data(
        request: Request,
        response: ApiExceptionResponse,
        start_time: float,
        end_time: float | None,
    ) -> RequestLogDict:
        """
        解析请求记录
        :param request: 请求
        :param response: 响应
        :param start_time: 开始时间
        :param end_time: 结束时间
        :return:
        """
        # region 获取用户
        jwt = request.headers.get("authorization")
        try:
            if jwt:  # 有jwt，解析
                payload = jwt.split(" ")[1].split(".")[1]
                payload = json.loads(
                    base64.b64decode(
                        payload + "=" * (-len(payload) % 4)
                    ).decode()
                )
                user_id = payload.get(api_settings.USER_ID_CLAIM, None)
            else:  # 无jwt，使用request内的用户
                user_id = (
                    request.user.id if request.user.is_authenticated else None
                )
        except Exception:
            user_id = None
        # endregion

        # region 记录请求参数
        request_param = request.GET.dict()

        request_data: Dict[str, JSONVal] = {}
        file_data: Dict[str, FileDataDict] = {}
        try:
            for key, value in response.api_request_data.items():
                if isinstance(value, UploadedFile):  # 文件
                    file_data[key]: FileDataDict = {
                        "name": value.name,
                        "size": value.size,
                        "content_type": value.content_type,
                        "content_type_extra": value.content_type_extra,
                    }
                else:  # 文本数据
                    request_data[key] = value
        except Exception:
            pass
        # endregion

        # region 记录响应数据
        response_body = {}
        try:
            if response.get("content-type") in [
                "application/json",
                "application/vnd.api+json",
            ]:  # 只记录json格式的响应
                if getattr(response, "streaming", False):  # 流式响应
                    response_body = {"__content__": "streaming"}
                else:  # 文本响应
                    if type(response.content) == bytes:  # bytes类型
                        response_body = json.loads(response.content.decode())
                    else:  # str类型
                        response_body = json.loads(response.content)
            elif "gzip" in response.get("content-type"):
                response_body = {"__content__": "gzip file"}
        except Exception:
            response_body = {"__content__": "parse error"}
        # endregion

        # region 记录url
        if drf_logger_settings.PATH_TYPE == "ABSOLUTE":
            url = request.build_absolute_uri()
        elif drf_logger_settings.PATH_TYPE == "FULL_PATH":
            url = request.get_full_path()
        else:
            url = request.build_absolute_uri()
        # endregion

        headers = get_headers(request=request)
        method = request.method

        return dict(
            user=user_id,
            ip=get_client_ip(request),
            method=method,
            url=url,
            headers=mask_sensitive_data(headers),
            content_type=request.content_type,
            query_param=mask_sensitive_data(request_param),
            request_body=mask_sensitive_data(request_data),
            file_data=file_data,
            response=mask_sensitive_data(response_body),
            status_code=response.status_code,
            execution_time=end_time - start_time
            if start_time and end_time
            else None,
            create_time=timezone.now(),
        )



import asyncio
import time

from zq_django_util.logs.threads import LOGGER_THREAD
from zq_django_util.logs.utils import database_log_enabled


class APILoggerMiddleware:
    sync_capable = True
    async_capable = True

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.
        self._is_coroutine = asyncio.iscoroutinefunction(get_response)

    def __call__(self, request):
        if not database_log_enabled():
            return self.get_response(request)

        if self._is_coroutine:
            return self.insert_log_async(request)
        else:
            return self.insert_log(request)

    def insert_log(self, request):
        start_time = time.time()
        response = self.get_response(request)
        end_time = time.time()
        LOGGER_THREAD.put_log_data(request, response, start_time, end_time)

        return response

    async def insert_log_async(self, request):
        start_time = time.time()
        response = await self.get_response(request)
        end_time = time.time()
        LOGGER_THREAD.put_log_data(request, response, start_time, end_time)

        return response



from typing import List, TypedDict

from django.core.signals import setting_changed
from django.dispatch import receiver

from zq_django_util.utils.package_settings import PackageSettings

OssSettingDict = TypedDict(
    "OssSettingDict",
    {
        "ACCESS_KEY_ID": str,
        "ACCESS_KEY_SECRET": str,
        "END_POINT": str,
        "BUCKET_NAME": str,
        "URL_EXPIRE_SECOND": int,
        "TOKEN_EXPIRE_SECOND": int,
        "MAX_SIZE_MB": int,
    },
)


class OssSettings(PackageSettings):
    setting_name = "ALIYUN_OSS"

    DEFAULTS: OssSettingDict = {
        "ACCESS_KEY_ID": "",
        "ACCESS_KEY_SECRET": "",
        "ENDPOINT": "https://oss-cn-shanghai.aliyuncs.com",
        "BUCKET_NAME": "",
        "URL_EXPIRE_SECOND": 60 * 60 * 24 * 30,
        "TOKEN_EXPIRE_SECOND": 60,
        "MAX_SIZE_MB": 100,
    }

    IMPORT_STRINGS: List[str] = []


oss_settings = OssSettings()


@receiver(setting_changed)
def reload_settings(*args, **kwargs):
    oss_settings.reload_package_settings(*args, **kwargs)



class AuthException(Exception):
    def __init__(self, message):
        super().__init__(message)


class OpenIdException(AuthException):
    def __init__(self, message):
        super().__init__(message)


class OpenIdNotBound(OpenIdException):
    def __init__(self, openid: str = ""):
        super().__init__(f"OpenID {openid} not bound")


class OpenIdNotProvided(OpenIdException):
    def __init__(self):
        super().__init__("OpenID not provided")



import csv
from datetime import timedelta
from typing import TYPE_CHECKING

from django.contrib import admin
from django.db.models import Count
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _

from zq_django_util.logs import models
from zq_django_util.logs.configs import drf_logger_settings

if TYPE_CHECKING:
    from zq_django_util.logs.models import RequestLog


@admin.register(models.ExceptionLog)
class ExceptionLogAdmin(admin.ModelAdmin):
    list_per_page = 20  # 每页显示条数
    list_display = [
        "exp_id",
        "exception_type",
        "method",
        "url",
        "ip",
        "user",
        "create_time",
    ]
    list_display_links = ["exp_id", "exception_type"]
    search_fields = ["exp_id", "exception_type", "ip", "url", "user"]
    list_filter = ["exception_type", "method", "url", "user"]
    ordering = ["-id"]

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class ExportCsvMixin:
    def export_as_csv(self, request, queryset):
        meta = self.model._meta
        field_names = [field.name for field in meta.fields]

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={}.csv".format(
            meta
        )
        writer = csv.writer(response)

        writer.writerow(field_names)
        for obj in queryset:
            writer.writerow([getattr(obj, field) for field in field_names])

        return response

    export_as_csv.short_description = "Export Selected"


class SlowAPIsFilter(admin.SimpleListFilter):
    title = _("API Performance")

    # Parameter for the filter that will be used in the URL query.
    parameter_name = "api_performance"

    def __init__(self, request, params, model, model_admin):
        super().__init__(request, params, model, model_admin)
        if (
            type(drf_logger_settings.ADMIN_SLOW_API_ABOVE) == int
        ):  # Making sure for integer value.
            # Converting to seconds.
            self._DRF_API_LOGGER_SLOW_API_ABOVE = (
                drf_logger_settings.ADMIN_SLOW_API_ABOVE / 1000
            )
        else:
            raise ValueError(
                "DRF_LOGGER__ADMIN_SLOW_API_ABOVE must be an integer."
            )

    def lookups(self, request, model_admin):
        """
        Returns a list of tuples. The first element in each
        tuple is the coded value for the option that will
        appear in the URL query. The second element is the
        human-readable name for the option that will appear
        in the right sidebar.
        """
        slow = "Slow, >={}ms".format(drf_logger_settings.ADMIN_SLOW_API_ABOVE)
        fast = "Fast, <{}ms".format(drf_logger_settings.ADMIN_SLOW_API_ABOVE)

        return (
            ("slow", _(slow)),
            ("fast", _(fast)),
        )

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via
        `self.value()`.
        """
        # to decide how to filter the queryset.
        if self.value() == "slow":
            return queryset.filter(
                execution_time__gte=self._DRF_API_LOGGER_SLOW_API_ABOVE
            )
        if self.value() == "fast":
            return queryset.filter(
                execution_time__lt=self._DRF_API_LOGGER_SLOW_API_ABOVE
            )

        return queryset


@admin.register(models.RequestLog)
class RequestLogAdmin(admin.ModelAdmin, ExportCsvMixin):
    actions = ["export_as_csv"]

    def __init__(self, model, admin_site):
        super().__init__(model, admin_site)
        self._DRF_API_LOGGER_TIMEDELTA = 0

        self.list_filter += (SlowAPIsFilter,)
        if (
            type(drf_logger_settings.ADMIN_TIMEDELTA) == int
        ):  # Making sure for integer value.
            self._DRF_API_LOGGER_TIMEDELTA = drf_logger_settings.ADMIN_TIMEDELTA

    def added_on_time(self, obj: "RequestLog") -> str:
        return (
            obj.create_time + timedelta(minutes=self._DRF_API_LOGGER_TIMEDELTA)
        ).strftime("%d %b %Y %H:%M:%S")

    added_on_time.admin_order_field = "create_time"
    added_on_time.short_description = "Create at"

    list_per_page = 20
    list_display = ["method", "url", "status_code", "ip", "user", "create_time"]
    list_display_links = ["method", "url"]
    search_fields = ["ip", "url", "user"]
    list_filter = ["method", "url", "status_code", "user"]
    ordering = ["-id"]

    change_list_template = "charts_change_list.html"
    change_form_template = "change_form.html"
    date_hierarchy = "create_time"

    def changelist_view(self, request, extra_context=None):
        response = super(RequestLogAdmin, self).changelist_view(
            request, extra_context
        )
        try:
            filtered_query_set = response.context_data["cl"].queryset
        except Exception:
            return response
        analytics_model = (
            filtered_query_set.values("create_time__date")
            .annotate(total=Count("id"))
            .order_by("total")
        )
        status_code_count_mode = (
            filtered_query_set.values("id")
            .values("status_code")
            .annotate(total=Count("id"))
            .order_by("status_code")
        )
        status_code_count_keys = list()
        status_code_count_values = list()
        for item in status_code_count_mode:
            status_code_count_keys.append(item.get("status_code"))
            status_code_count_values.append(item.get("total"))
        extra_context = dict(
            analytics=analytics_model,
            status_code_count_keys=status_code_count_keys,
            status_code_count_values=status_code_count_values,
        )
        response.context_data.update(extra_context)
        return response

    def get_queryset(self, request):
        return (
            super(RequestLogAdmin, self)
            .get_queryset(request)
            .using(drf_logger_settings.DEFAULT_DATABASE)
        )

    def changeform_view(
        self, request, object_id=None, form_url="", extra_context=None
    ):
        if request.GET.get("export", False):
            export_queryset = (
                self.get_queryset(request)
                .filter(pk=object_id)
                .using(drf_logger_settings.DEFAULT_DATABASE)
            )
            return self.export_as_csv(request, export_queryset)
        return super(RequestLogAdmin, self).changeform_view(
            request, object_id, form_url, extra_context
        )

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False



# 自定义返回格式
from typing import Any, Mapping, Optional

from rest_framework.renderers import JSONRenderer

from zq_django_util.response import ApiResponse
from zq_django_util.response.types import ApiExceptionResponse


class CustomRenderer(JSONRenderer):
    # 重构render方法
    def render(
        self,
        data: Any,
        accepted_media_type: Optional[str] = None,
        renderer_context: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        if renderer_context:
            response: ApiExceptionResponse = renderer_context["response"]
            try:  # 记录请求数据，便于日志处理
                request_data = renderer_context["request"].data
                response.api_request_data = request_data
            except Exception:
                pass

            if not response.exception:  # 如果不是异常
                data = ApiResponse(data=data).__dict__()  # 将data包装成ApiResponse

        return super().render(data, accepted_media_type, renderer_context)



from django.http import HttpRequest, JsonResponse

from zq_django_util.response import ApiResponse, ResponseType


def bad_request(request: HttpRequest, exception: Exception) -> JsonResponse:
    """
    404 页面

    在 url.py 中设定 handler404 为此函数

    """
    return ApiResponse(ResponseType.APINotFound, "您访问的页面不存在").to_json_response()


def server_error(request: HttpRequest) -> JsonResponse:
    """
    500 页面

    在 url.py 中设定 handler500 为此函数

    """
    # TODO 结合异常处理(记录日志)
    return ApiResponse(ResponseType.ServerError).to_json_response()



# 重写 jwt 相关验证类
from typing import Any, Dict, Optional

from django.contrib.auth import get_user_model
from rest_framework_simplejwt.authentication import JWTAuthentication
from rest_framework_simplejwt.settings import api_settings

from zq_django_util.exceptions import ApiException
from zq_django_util.response import ResponseType


class ActiveUserAuthentication(JWTAuthentication):
    """
    激活用户认证类
    """

    AuthUser = get_user_model()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.user_model = self.AuthUser

    def get_user(self, validated_token: Dict[str, Any]) -> Optional[AuthUser]:
        """
        Attempts to find and return a user using the given validated token.
        """
        # 获取用户 ID
        try:
            user_id = validated_token[api_settings.USER_ID_CLAIM]
        except KeyError:
            raise ApiException(
                ResponseType.TokenInvalid, "token获取失败", record=True
            )  # token 不存在, 认证失败

        # 获取用户模型
        try:
            user = self.user_model.objects.get(
                **{api_settings.USER_ID_FIELD: user_id}
            )
        except self.user_model.DoesNotExist:
            raise ApiException(
                ResponseType.ResourceNotFound, "用户不存在", record=True
            )  # 用户不存在, 认证失败

        self.check_activate(user)  # 检查用户激活类型

        return user

    @staticmethod
    def check_activate(user: AuthUser) -> None:
        """
        检查是否激活
        """
        if not user.is_active:
            raise ApiException(ResponseType.NotActive, "您的账号未激活，请完善个人信息后重试")


class NormalUserAuthentication(ActiveUserAuthentication):
    """
    无需激活的用户认证类
    """

    AuthUser = get_user_model()

    @staticmethod
    def check_activate(user: AuthUser) -> None:
        pass



# backend 用于用户验证
# 由于项目不使用django认证权限管理，backend不通过django的认证模块调用
# https://docs.djangoproject.com/zh-hans/4.1/topics/auth/customizing/

from typing import Any, Optional

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import BaseBackend
from rest_framework.request import Request

from zq_django_util.utils.auth.exceptions import (
    OpenIdNotBound,
    OpenIdNotProvided,
)


class OpenIdBackend(BaseBackend):
    """
    OpenID 验证模块(获取token时调用)
    """

    AuthUser = get_user_model()

    openid_field: str = getattr(
        settings, "OPENID_FIELD", "openid"
    )  # 获取 openid 字段名

    def __init__(self, auth_user_model=None):
        super().__init__()
        if auth_user_model:
            self.AuthUser = auth_user_model

    def authenticate(
        self,
        request: Optional[Request] = None,
        raise_exception: bool = False,
        **kwargs: Any,
    ) -> Optional[AuthUser]:
        """
        重写认证方法

        :param raise_exception: 是否产生异常，若为否则兼容django认证模块

        :param request: 传入请求


        :param kwargs: 字典参数

        :return: openid 对应用户对象
        """
        openid = kwargs.get(self.openid_field, None)

        try:
            # 未传入openid 认证失败
            if openid is None:
                raise OpenIdNotProvided
            # 获取 openid 对应用户
            try:
                return self.AuthUser.objects.get(**{self.openid_field: openid})
            except self.AuthUser.DoesNotExist:
                raise OpenIdNotBound(openid)
        except (OpenIdNotProvided, OpenIdNotBound) as e:
            if raise_exception:
                raise e
            return None

    def get_user(self, user_id: int) -> Optional[AuthUser]:
        """
        重写用户获取方法

        :param user_id: uid

        :return: 用户对象
        """
        try:
            user = self.AuthUser.objects.get(pk=user_id)
        except self.AuthUser.DoesNotExist:  # 用户不存在, 返回 None
            return None
        return user



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zq_django_util',
 'zq_django_util.exceptions',
 'zq_django_util.logs',
 'zq_django_util.logs.migrations',
 'zq_django_util.response',
 'zq_django_util.utils',
 'zq_django_util.utils.auth',
 'zq_django_util.utils.oss',
 'zq_django_util.utils.user']

package_data = \
{'': ['*'], 'zq_django_util.logs': ['templates/*']}

install_requires = \
['django-cleanup>=6.0.0,<7.0.0',
 'djangorestframework-simplejwt>=4.7,<6.0',
 'drf-standardized-errors>=0.9.0,<0.13.0',
 'oss2>=2.13.0,<3.0']

setup_kwargs = {
    'name': 'zq-django-util',
    'version': '0.1.10',
    'description': '自强Studio Django 工具',
    'long_description': '<div align="center">\n\n# zq-django-util\n**自强 Studio Django 工具**\n\n<!-- markdownlint-disable-next-line MD036 -->\n</div>\n\n<p align="center">\n  <a href="https://github.com/Nagico/zq-django-util/actions/workflows/code_check.yml">\n    <img src="https://github.com/Nagico/zq-django-util/actions/workflows/code_check.yml/badge.svg" alt="CI">\n  </a>\n  <a href="https://zq-django-util.readthedocs.io/en/latest/?badge=latest">\n    <img src="https://readthedocs.org/projects/zq-django-util/badge/?version=latest" alt="Documentation Status" />\n  </a>\n  <a href="https://codecov.io/gh/Nagico/zq-django-util" >\n    <img src="https://codecov.io/gh/Nagico/zq-django-util/branch/master/graph/badge.svg" alt="cov"/>\n  </a>\n  <a href="https://pypi.org/project/zq-django-util/">\n  <img src="https://img.shields.io/pypi/v/zq-django-util" alt="pypi">\n  </a>\n</p>\n<!-- markdownlint-enable MD033 -->\n\n[English Version](README_EN.md)\n\n## 简介\n\nzq-django-util 是用于辅助搭建 django-drf 应用的工具集合，其中包含：\n\n- 标准异常、响应处理\n- jwt、微信认证\n- oss 存储与直传\n- 默认分页类\n- 测试 ViewSet\n\n详细文档：[zq-django-util.readthedocs.io](https://zq-django-util.readthedocs.io/)\n\n## 依赖需求\n\n- Python 3.8+\n- Django 3.2+\n- Django REST framework 3.12+\n\n**强烈建议**使用官方支持的最新版本，当前的测试环境为：Python 3.10, Django 4.1, DRF 3.14\n\n## 安装\n\n- 安装 zq-django-util 包\n\n使用 `pip` 安装：\n```shell\npip install zq-django-util\n```\n\n使用 `poetry` 安装：\n```shell\npoetry add zq-django-util\n```\n\n## 使用\n\n可以根据以下说明进行配置，以启用相关功能。\n\n[使用文档](docs/usage)\n\n## 开发\n\n本项目使用 Poetry 进行依赖管理，Pytest 进行单元测试。\n\n[开发文档](docs/development)\n',
    'author': 'Nagico',
    'author_email': 'yjr888@vip.qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Nagico/zq-django-util',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import logging
import os
import shutil
from datetime import datetime, timezone
from tempfile import SpooledTemporaryFile
from typing import BinaryIO, List, Optional, Union
from urllib.parse import urljoin

import oss2
import oss2.exceptions
import oss2.utils
from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.core.files import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_str
from oss2.models import GetObjectMetaResult

from .configs import oss_settings
from .exceptions import OssError

logger = logging.getLogger("oss")


@deconstructible
class OssStorage(Storage):
    """
    Aliyun OSS Storage 基本存储
    """

    access_key_id: str
    access_key_secret: str
    end_point: str
    bucket_name: str
    expire_time: int

    auth: oss2.Auth
    service: oss2.Service
    bucket: oss2.Bucket
    bucket_acl: str

    base_dir: str  # 基本路径

    def __init__(
        self,
        access_key_id: Optional[str] = None,
        access_key_secret: Optional[str] = None,
        end_point: Optional[str] = None,
        bucket_name: Optional[str] = None,
        expire_time: Optional[int] = None,
    ):
        self.access_key_id = access_key_id or oss_settings.ACCESS_KEY_ID
        self.access_key_secret = (
            access_key_secret or oss_settings.ACCESS_KEY_SECRET
        )
        self.end_point = self._normalize_endpoint(
            end_point or oss_settings.ENDPOINT
        )
        self.bucket_name = bucket_name or oss_settings.BUCKET_NAME
        self.expire_time = expire_time or oss_settings.URL_EXPIRE_SECOND

        self.auth = oss2.Auth(self.access_key_id, self.access_key_secret)
        self.service = oss2.Service(self.auth, self.end_point)
        self.bucket = oss2.Bucket(self.auth, self.end_point, self.bucket_name)

        # try to get bucket acl to check bucket exist or not
        try:
            self.bucket_acl = self.bucket.get_bucket_acl().acl
        except oss2.exceptions.NoSuchBucket:
            raise SuspiciousOperation(
                "Bucket '%s' does not exist." % self.bucket_name
            )

    @staticmethod
    def _normalize_endpoint(endpoint: str) -> str:
        if not endpoint.startswith("http://") and not endpoint.startswith(
            "https://"
        ):
            return "http://" + endpoint
        else:
            return endpoint

    def _get_key_name(self, name: str) -> str:
        """
        Get the object key name in OSS, e.g.,
        base_dir: /media/
        :param name: test.txt
        :return: media/test.txt
        """
        # urljoin won't work if name is absolute path
        name = name.lstrip("/")

        base_path = force_str(self.base_dir)
        final_path = urljoin(base_path + "/", name)
        name = os.path.normpath(final_path.lstrip("/"))

        # Add / to the end of path since os.path.normpath will remove it
        if final_path.endswith("/") and not name.endswith("/"):
            name += "/"

        # Store filenames with forward slashes, even on Windows.
        return name.replace("\\", "/")

    def _open(self, name: str, mode: str = "rb") -> "OssFile":
        """
        Open a file for reading from OSS.
        :param name: 文件名
        :param mode: 打开模式
        :return:
        """
        logger.debug("name: %s, mode: %s", name, mode)
        if mode != "rb":
            raise ValueError("OSS files can only be opened in read-only mode")

        target_name = self._get_key_name(name)
        logger.debug("target name: %s", target_name)
        try:
            # Load the key into a temporary file
            tmp_file = SpooledTemporaryFile(max_size=10 * 1024 * 1024)  # 10MB
            obj = self.bucket.get_object(target_name)
            logger.info(
                "content length: %d, requestid: %s",
                obj.content_length,
                obj.request_id,
            )
            if obj.content_length is None:
                shutil.copyfileobj(obj, tmp_file)
            else:
                oss2.utils.copyfileobj_and_verify(
                    obj, tmp_file, obj.content_length, request_id=obj.request_id
                )
            tmp_file.seek(0)
            return OssFile(tmp_file, target_name, self)
        except oss2.exceptions.NoSuchKey:
            raise OssError("%s does not exist" % name)
        except Exception:
            raise OssError("Failed to open %s" % name)

    def _save(self, name: str, content: Union[File, bytes, str]) -> str:
        target_name = self._get_key_name(name)
        logger.debug("target name: %s", target_name)
        logger.debug("content: %s", content)
        self.bucket.put_object(target_name, content)
        return os.path.normpath(name)

    def create_dir(self, dirname: str) -> None:
        """
        创建目录
        :param dirname: 文件夹路径
        :return:
        """
        target_name = self._get_key_name(dirname)
        if not target_name.endswith("/"):
            target_name += "/"

        self.bucket.put_object(target_name, "")

    def exists(self, name: str) -> bool:
        target_name = self._get_key_name(name)
        logger.debug("name: %s, target name: %s", name, target_name)
        return self.bucket.object_exists(target_name)

    def get_file_meta(self, name: str) -> GetObjectMetaResult:
        """
        获取文件信息
        :param name: 文件名
        :return:
        """
        name = self._get_key_name(name)
        return self.bucket.get_object_meta(name)

    def size(self, name: str) -> int:
        file_meta = self.get_file_meta(name)
        return file_meta.content_length

    def modified_time(self, name: str) -> datetime:
        file_meta = self.get_file_meta(name)
        return datetime.fromtimestamp(file_meta.last_modified)

    created_time = accessed_time = modified_time

    def get_modified_time(self, name: str) -> datetime:
        file_meta = self.get_file_meta(name)

        if settings.USE_TZ:
            return datetime.utcfromtimestamp(file_meta.last_modified).replace(
                tzinfo=timezone.utc
            )
        else:
            return datetime.fromtimestamp(file_meta.last_modified)

    get_created_time = get_accessed_time = get_modified_time

    def content_type(self, name: str) -> str:
        name = self._get_key_name(name)
        file_info = self.bucket.head_object(name)
        return file_info.content_type

    def listdir(self, name: str) -> (List[str], List[str]):
        if name == ".":
            name = ""
        name = self._get_key_name(name)
        if not name.endswith("/"):
            name += "/"
        logger.debug("name: %s", name)

        files: list[str] = []
        dirs: list[str] = []

        for obj in oss2.ObjectIterator(self.bucket, prefix=name, delimiter="/"):
            if obj.is_prefix():
                dirs.append(obj.key)
            else:
                files.append(obj.key)

        logger.debug("dirs: %s", list(dirs))
        logger.debug("files: %s", files)
        return dirs, files

    def url(self, name: str) -> str:
        """
        获取文件的url(带token)
        :param name: 文件名
        :return: url
        """
        key = self._get_key_name(name)
        url = self.bucket.sign_url("GET", key, expires=self.expire_time)
        if self.bucket_acl != oss2.BUCKET_ACL_PRIVATE:
            idx = url.find("?")
            if idx > 0:
                url = url[:idx].replace("%2F", "/")
        return url

    def delete(self, name: str) -> None:
        """
        删除文件

        :param name:
        :return:
        """
        name = self._get_key_name(name)
        logger.debug("delete name: %s", name)
        self.bucket.delete_object(name)

    def delete_dir(self, dirname: str) -> None:
        """
        删除文件夹

        :param dirname:
        :return:
        """
        name = self._get_key_name(dirname)
        if not name.endswith("/"):
            name += "/"
        logger.debug("delete name: %s", name)
        self.bucket.delete_object(name)

    def get_object_acl(self, name: str) -> str:
        """
        获取文件的访问权限
        :param name: 文件名
        :return:
        """
        name = self._get_key_name(name)
        return self.bucket.get_object_acl(name).acl

    def set_object_acl(self, name: str, acl: str) -> None:
        """
        设置文件的访问权限
        :param name: 文件名
        :param acl: 访问权限
        :return:
        """
        name = self._get_key_name(name)
        self.bucket.put_object_acl(name, acl)


class OssMediaStorage(OssStorage):
    def __init__(self):
        self.base_dir = settings.MEDIA_URL
        logger.debug("locatin: %s", self.base_dir)
        super(OssMediaStorage, self).__init__()


class OssStaticStorage(OssStorage):
    def __init__(self):
        self.base_dir = settings.STATIC_URL
        logger.info("locatin: %s", self.base_dir)
        super(OssStaticStorage, self).__init__()


class OssFile(File):
    """
    A file returned from AliCloud OSS
    """

    def __init__(
        self, content: SpooledTemporaryFile, name: str, storage: OssStorage
    ):
        super(OssFile, self).__init__(content, name)
        self._storage = storage

    def open(self, mode: str = "rb") -> BinaryIO:
        if self.closed:
            self.file = self._storage.open(self.name, mode).file
        return super(OssFile, self).open(mode)



from django.contrib.auth.models import AbstractUser as DjangoAbstractUser
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.db import models


class AbstractUser(DjangoAbstractUser):
    """
    基本用户表
    """

    # 默认字段 - 基本信息
    username = models.CharField(
        "用户名",
        max_length=150,
        unique=True,
        validators=[UnicodeUsernameValidator()],
    )
    password = models.CharField("密码", max_length=128)
    email = models.EmailField("邮箱", blank=True)

    # 默认字段 - 权限
    is_staff = models.BooleanField("工作人员", default=False)
    is_active = models.BooleanField("是否激活", default=True)
    is_superuser = models.BooleanField("超级用户", default=False)

    # 自定义字段
    # openid = models.CharField(
    #     max_length=64, unique=True, verbose_name="微信openid"
    # )

    avatar = models.ImageField(
        upload_to="avatar", default=r"avatar\default.jpg", verbose_name="头像"
    )

    phone = models.CharField(max_length=13, default="", verbose_name="手机")

    create_time = models.DateTimeField(auto_now_add=True, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    # 禁用的默认字段
    first_name = None
    last_name = None
    date_joined = None
    last_login = None

    def __str__(self):
        return self.username

    class Meta:
        abstract = True
        db_table = "zq_user"
        verbose_name = "用户"
        verbose_name_plural = verbose_name
        ordering = ["-id"]



from datetime import datetime
from typing import Any, Dict

import rest_framework_simplejwt.settings
from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework import serializers
from rest_framework_simplejwt.serializers import (
    PasswordField,
    TokenObtainPairSerializer,
)
from rest_framework_simplejwt.tokens import RefreshToken

from zq_django_util.exceptions import ApiException
from zq_django_util.response import ResponseType
from zq_django_util.utils.auth.backends import OpenIdBackend
from zq_django_util.utils.auth.exceptions import OpenIdNotBound

AuthUser = get_user_model()


class OpenIdLoginSerializer(serializers.Serializer):
    """
    OpenID Token 获取序列化器
    """

    openid_field: str = getattr(settings, "OPENID_FIELD", "openid")
    backend = OpenIdBackend()

    openid = PasswordField()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def get_token(cls, user: AuthUser) -> RefreshToken:
        """
        获取 token(access与refresh)

        :param user: user对象

        :return: token 对象
        """
        # 调用 simple-jwt 中 token 生成方法, 需要在 settings 中指定 USER_ID_FIELD 为主键名
        return RefreshToken.for_user(user)

    def validate(self, attrs: Dict[str, Any]) -> dict:
        """
        重写验证器

        :param attrs: 序列化器中待验证的数据

        :return: 已验证数据，返回前端
        """
        openid = self.get_open_id(attrs)
        # 验证 openid
        authenticate_kwargs = {
            self.openid_field: openid
        }  # 给 openid 验证模块准备 openid
        try:
            #  给 openid 验证模块准备请求数据
            authenticate_kwargs["request"] = self.context["request"]
        except KeyError:
            pass

        openid_backend = self.backend

        try:
            user: AuthUser = openid_backend.authenticate(
                **authenticate_kwargs, raise_exception=True
            )  # 调用 openid 验证模块进行权限验证
        except OpenIdNotBound:  # openid 未绑定用户
            user = self.handle_new_openid(openid)  # 处理新 openid

        # token 获取
        user_id_field = (
            rest_framework_simplejwt.settings.api_settings.USER_ID_FIELD
        )  # 读取 settings 中定义的 user 主键字段名
        refresh = self.get_token(user)  # 获取 token

        return self.generate_token_result(
            user,
            user_id_field,
            refresh.access_token.current_time + refresh.access_token.lifetime,
            str(refresh.access_token),
            str(refresh),
        )

    def get_open_id(self, attrs: Dict[str, Any]) -> str:
        """
        获取 openid
        """
        return attrs[self.openid_field]

    def generate_token_result(
        self,
        user: AuthUser,
        user_id_field: str,
        expire_time: datetime,
        access: str,
        refresh: str,
    ) -> dict:
        """
        生成 token 返回结果
        :param user: 用户对象
        :param user_id_field: 用户主键字段
        :param expire_time: 过期时间
        :param access: access token
        :param refresh: refresh token
        :return:
        """
        return dict(
            id=getattr(user, user_id_field),
            username=user.username,
            expire_time=expire_time,
            access=access,
            refresh=refresh,
        )

    def handle_new_openid(self, openid: str) -> AuthUser:
        """
        处理新 openid
        如果自动注册需要返回对象
        """
        raise ApiException(
            ResponseType.ThirdLoginFailed,
            msg="该微信账号暂未未绑定用户",
            detail="openid未绑定",
        )


class AbstractWechatLoginSerializer(OpenIdLoginSerializer):
    """
    微信登录序列化器
    """

    code = PasswordField(label="前端获取code")  # 前端传入 code

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields.pop("openid")  # 删除 openid 字段

    def get_open_id(self, attrs: Dict[str, Any]) -> str:
        """
        重写获取 open_id 方法
        """
        raise NotImplementedError("请在此返回openid")


class PasswordLoginSerializer(TokenObtainPairSerializer):
    def validate(self, attrs: Dict[str, Any]) -> dict:
        super(TokenObtainPairSerializer, self).validate(attrs)  # 获取 self.user

        user_id_field = (
            rest_framework_simplejwt.settings.api_settings.USER_ID_FIELD
        )  # 读取 settings 中定义的 user 主键字段名

        refresh = self.get_token(self.user)

        return self.generate_token_result(
            self.user,
            user_id_field,
            refresh.access_token.current_time + refresh.access_token.lifetime,
            str(refresh.access_token),
            str(refresh),
        )

    def generate_token_result(
        self,
        user: AuthUser,
        user_id_field: str,
        expire_time: datetime,
        access: str,
        refresh: str,
    ) -> dict:
        """
        生成 token 返回结果
        :param user: 用户对象
        :param user_id_field: 用户主键字段
        :param expire_time: 过期时间
        :param access: access token
        :param refresh: refresh token
        :return:
        """
        return dict(
            id=getattr(user, user_id_field),
            username=user.username,
            expire_time=expire_time,
            access=access,
            refresh=refresh,
        )



# Generated by Django 4.1.7 on 2023-03-01 11:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("logs", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="requestlog",
            name="content_type",
            field=models.CharField(max_length=34, verbose_name="请求类型"),
        ),
        migrations.AlterField(
            model_name="requestlog",
            name="execution_time",
            field=models.DecimalField(
                decimal_places=8, max_digits=10, null=True, verbose_name="执行时间"
            ),
        ),
        migrations.AlterField(
            model_name="requestlog",
            name="ip",
            field=models.CharField(max_length=16, verbose_name="用户IP"),
        ),
        migrations.AlterField(
            model_name="requestlog",
            name="method",
            field=models.CharField(max_length=8, verbose_name="请求方法"),
        ),
    ]



from django.db import models


class RequestLog(models.Model):
    """
    Model to store request logs
    """

    user = models.IntegerField(null=True, blank=True, verbose_name="用户ID")
    ip = models.CharField(max_length=16, verbose_name="用户IP")
    method = models.CharField(max_length=8, verbose_name="请求方法")
    url = models.TextField(verbose_name="请求URL")
    headers = models.JSONField(verbose_name="请求头")
    content_type = models.CharField(max_length=34, verbose_name="请求类型")
    query_param = models.JSONField(verbose_name="请求参数")
    request_body = models.JSONField(verbose_name="请求数据")
    file_data = models.JSONField(verbose_name="文件数据")

    response = models.JSONField(verbose_name="响应数据")

    status_code = models.PositiveSmallIntegerField(
        db_index=True, verbose_name="响应状态码"
    )
    execution_time = models.DecimalField(
        null=True, decimal_places=8, max_digits=10, verbose_name="执行时间"
    )
    create_time = models.DateTimeField(auto_now_add=True, verbose_name="请求时间")

    def __str__(self):
        return self.ip

    class Meta:
        ordering = ["-create_time"]
        app_label = "logs"
        db_table = "log_request"
        verbose_name = "请求日志"
        verbose_name_plural = verbose_name


class ExceptionLog(RequestLog):
    """
    Model to store exception logs
    """

    exp_id = models.CharField(max_length=32, verbose_name="异常ID")
    exception_type = models.CharField(max_length=128, verbose_name="异常类型")
    event_id = models.CharField(max_length=32, verbose_name="Sentry事件ID")
    exception_msg = models.TextField(verbose_name="异常信息")
    exception_info = models.TextField(verbose_name="异常详情")
    stack_info = models.JSONField(verbose_name="异常栈")

    def __str__(self):
        return self.exception_type

    class Meta:
        ordering = ["-create_time"]
        app_label = "logs"
        db_table = "log_exception"
        verbose_name = "异常日志"
        verbose_name_plural = verbose_name
